from datetime import datetime

from typing import Optional

from example.messages.ender_after_answer_message import EnderAfterAnswerMessage
from example.messages.ender_after_input_message import EnderAfterInputMessage
from example.messages.input_message import InputMessage
from example.messages.some_external_message import SomeExternalMessage
from unipipeline.answer.uni_answer_message import UniAnswerMessage
from unipipeline.worker.uni_msg_params import UniSendingParams
from unipipeline.worker.uni_worker import UniWorker
from unipipeline.worker.uni_worker_consumer_message import UniWorkerConsumerMessage


class InputWorker(UniWorker[InputMessage, None]):
    def handle_message(self, msg: UniWorkerConsumerMessage[InputMessage]) -> None:
        # answ = '!!!'  # * 100000

        self.manager.send_to('ender_after_input_worker', EnderAfterInputMessage(
            value=f'from input_worker {datetime.now()}'
        ))

        self.manager.send_to('ender_after_input_worker', EnderAfterInputMessage(
            value=f'from input_worker {datetime.now()}'
        ))

        answ_msg: Optional[UniAnswerMessage[EnderAfterAnswerMessage]] = self.manager.get_answer_from('ender_after_input_worker', EnderAfterInputMessage(value=f'from input_worker {datetime.now()}'))
        assert answ_msg is not None
        answ = answ_msg.payload.value

        self.manager.send_to('some_external_worker', SomeExternalMessage(
            value=f'answ: {answ[:30]} ==> from input_worker {datetime.now()}'
        ), UniSendingParams(ttl_s=15))



from datetime import timedelta
from typing import Union, Dict, Any, List, Tuple, Type, TYPE_CHECKING, Optional

from pydantic import BaseModel

from unipipeline.message.uni_message import UniMessage

if TYPE_CHECKING:
    from unipipeline.worker.uni_worker import UniWorker


TUniSendingWorkerUnion = Union[Type['UniWorker[UniMessage, UniMessage]'], str]
TUniSendingMessagePayloadUnion = Union[Dict[str, Any], UniMessage, List[Dict[str, Any]], Tuple[Dict[str, Any], ...], List[UniMessage], Tuple[UniMessage, ...]]


class UniSendingParams(BaseModel):
    alone: bool = False
    ttl_s: Optional[int] = None


class UniGettingAnswerParams(BaseModel):
    answer_tll: timedelta
    alone: bool = False


default_sending_params: UniSendingParams = UniSendingParams(alone=False, ttl_s=None)
default_getting_answer_params: UniGettingAnswerParams = UniGettingAnswerParams(answer_tll=timedelta(seconds=5))



from typing import Dict, Any, Set, Union, Type, Iterator, Tuple, Optional
from uuid import uuid4

import yaml

from unipipeline.definitions.uni_broker_definition import UniBrokerDefinition
from unipipeline.definitions.uni_codec_definition import UniCodecDefinition
from unipipeline.definitions.uni_cron_task_definition import UniCronTaskDefinition
from unipipeline.definitions.uni_external_definition import UniExternalDefinition
from unipipeline.definitions.uni_message_definition import UniMessageDefinition
from unipipeline.definitions.uni_module_definition import UniModuleDefinition
from unipipeline.definitions.uni_service_definition import UniServiceDefinition
from unipipeline.definitions.uni_waiting_definition import UniWaitingDefinition
from unipipeline.definitions.uni_worker_definition import UniWorkerDefinition
from unipipeline.errors import UniConfigError, UniDefinitionNotFoundError
from unipipeline.utils.filter_camel import camel_case
from unipipeline.utils.uni_echo import UniEcho
from unipipeline.utils.uni_util import UniUtil
from unipipeline.worker.uni_worker import UniWorker

UNI_CRON_MESSAGE = "uni_cron_message"


class UniConfig:
    # @staticmethod
    # def from_file(file_path: str) -> 'UniConfig':
    #     return UniConfig()

    def __init__(self, file_path: str, echo_level: Optional[Union[str, int]] = None) -> None:
        self._util = UniUtil()
        self._util.template.set_filter('camel', camel_case)

        self._file_path = file_path
        self._echo = UniEcho('UNI', level=echo_level, colors=self._util.color)

        self._config: Dict[str, Any] = dict()
        self._parsed = False
        self._config_loaded = False
        self._compression_index: Dict[str, UniCodecDefinition] = dict()
        self._codecs_index: Dict[str, UniCodecDefinition] = dict()
        self._waiting_index: Dict[str, UniWaitingDefinition] = dict()
        self._external: Dict[str, UniExternalDefinition] = dict()
        self._brokers_index: Dict[str, UniBrokerDefinition] = dict()
        self._messages_index: Dict[str, UniMessageDefinition] = dict()
        self._workers_by_name_index: Dict[str, UniWorkerDefinition] = dict()
        self._workers_by_class_index: Dict[str, UniWorkerDefinition] = dict()
        self._cron_tasks_index: Dict[str, UniCronTaskDefinition] = dict()
        self._service: UniServiceDefinition = None  # type: ignore

    @property
    def file(self) -> str:
        return self._file_path

    @property
    def brokers(self) -> Dict[str, UniBrokerDefinition]:
        self._parse()
        return self._brokers_index

    @property
    def service(self) -> UniServiceDefinition:
        self._parse()
        return self._service

    @property
    def compression(self) -> Dict[str, UniCodecDefinition]:
        self._parse()
        return self._compression_index

    @property
    def codecs(self) -> Dict[str, UniCodecDefinition]:
        self._parse()
        return self._codecs_index

    @property
    def cron_tasks(self) -> Dict[str, UniCronTaskDefinition]:
        self._parse()
        return self._cron_tasks_index

    @property
    def external(self) -> Dict[str, UniExternalDefinition]:
        self._parse()
        return self._external

    @property
    def workers(self) -> Dict[str, UniWorkerDefinition]:
        self._parse()
        return self._workers_by_name_index

    @property
    def workers_by_class(self) -> Dict[str, UniWorkerDefinition]:
        self._parse()
        return self._workers_by_class_index

    @property
    def waitings(self) -> Dict[str, UniWaitingDefinition]:
        self._parse()
        return self._waiting_index

    @property
    def messages(self) -> Dict[str, UniMessageDefinition]:
        self._parse()
        return self._messages_index

    def get_worker_definition(self, worker: Union[Type['UniWorker[Any, Any]'], str]) -> UniWorkerDefinition:
        if isinstance(worker, str):
            if worker in self.workers:
                return self.workers[worker]
            else:
                raise UniDefinitionNotFoundError(f'worker {worker} is not defined in workers')
        elif issubclass(worker, UniWorker):
            if worker.__name__ in self.workers_by_class:
                return self.workers_by_class[worker.__name__]
            else:
                raise UniDefinitionNotFoundError(f'worker {worker.__name__} is not defined in workers')
        raise UniDefinitionNotFoundError(f'invalid type of worker. must be subclass of UniWorker OR str name. "{type(worker).__name__}" was given')

    def _load_config(self) -> Dict[str, Any]:
        if self._config_loaded:
            return self._config
        self._config_loaded = True
        with open(self._file_path, "rt") as f:
            self._config = yaml.safe_load(f)
        if not isinstance(self._config, dict):
            raise UniConfigError('config must be dict')
        return self._config

    def _parse_definition(
        self,
        conf_name: str,
        definitions: Dict[str, Any],
        defaults: Dict[str, Any],
        required_keys: Set[str]
    ) -> Iterator[Tuple[str, Dict[str, Any], Dict[str, Any]]]:
        if not isinstance(definitions, dict):
            raise UniConfigError(f'worker_definition of {conf_name} has invalid type. must be dict')

        common = definitions.get("__default__", dict())

        for name, raw_definition in definitions.items():
            other_props: Dict[str, Any] = dict()

            name = str(name)
            if name.startswith("_"):
                if name == "__default__":
                    continue
                else:
                    raise ValueError(f"key '{name}' is not acceptable")

            if not isinstance(raw_definition, dict):
                raise UniConfigError(f'worker_definition of {conf_name}->{name} has invalid type. must be dict')

            result_definition = dict(defaults)
            combined_definition = dict(common)
            combined_definition.update(raw_definition)

            for k, v in combined_definition.items():
                if k in defaults:
                    result_definition[k] = v
                    vd = defaults.get(k, None)
                    if vd is not None and type(vd) != type(v):
                        raise UniConfigError(f'worker_definition of {conf_name}->{name} has invalid key "{k}" type')
                elif k in required_keys:
                    result_definition[k] = v
                else:
                    other_props[k] = v

            for rk in required_keys:
                if rk not in result_definition:
                    raise UniConfigError(f'worker_definition of {conf_name}->{name} has no required prop "{rk}"')

            result_definition["name"] = name
            result_definition["id"] = uuid4()

            yield name, result_definition, other_props

    def _parse(self) -> None:
        if self._parsed:
            return
        self._parsed = True

        cfg = self._load_config()

        self._service = self._parse_service(cfg)
        self._echo.log_info(f'service: {self._service.name}')

        self._compression_index = self._parse_compression(cfg)
        self._echo.log_info(f'compression codecs: {",".join(self._compression_index.keys())}')

        self._codecs_index = self._parse_codecs(cfg)
        self._echo.log_info(f'serialization codecs: {",".join(self._codecs_index.keys())}')

        self._external = self._parse_external_services(cfg)
        self._echo.log_info(f'external: {",".join(self._external.keys())}')

        self._waiting_index = self._parse_waitings(cfg, self._service)
        self._echo.log_info(f'waitings: {",".join(self._waiting_index.keys())}')

        self._brokers_index = self._parse_brokers(cfg, self._service, self._external)
        self._echo.log_info(f'brokers: {",".join(self._brokers_index.keys())}')

        self._messages_index = self._parse_messages(cfg, self._service)
        self._echo.log_info(f'messages: {",".join(self._messages_index.keys())}')

        self._workers_by_name_index = self._parse_workers(cfg, self._service, self._brokers_index, self._messages_index, self._waiting_index, self._external)
        self._echo.log_info(f'workers: {",".join(self._workers_by_name_index.keys())}')

        for wd in self._workers_by_name_index.values():
            if wd.marked_as_external:
                continue
            assert wd.type is not None
            self._workers_by_class_index[wd.type.object_name] = wd

        self._cron_tasks_index = self._parse_cron_tasks(cfg, self._service, self._workers_by_name_index)

    APPLICATION_JSON = 'application/json'
    COMPRESSION_GZIP = "application/x-gzip"
    COMPRESSION_BZ2 = "application/x-bz2"
    COMPRESSION_LZMA = "application/x-lzma"

    def _parse_compression(self, config: Dict[str, Any]) -> Dict[str, UniCodecDefinition]:
        result = {
            self.COMPRESSION_GZIP: UniCodecDefinition(
                name=self.COMPRESSION_GZIP,
                encoder_type=UniModuleDefinition.parse("zlib:compress"),
                decoder_type=UniModuleDefinition.parse("zlib:decompress"),
                dynamic_props_={}
            ),
            self.COMPRESSION_BZ2: UniCodecDefinition(
                name=self.COMPRESSION_BZ2,
                encoder_type=UniModuleDefinition.parse("bz2:compress"),
                decoder_type=UniModuleDefinition.parse("bz2:decompress"),
                dynamic_props_={}
            ),
            self.COMPRESSION_LZMA: UniCodecDefinition(
                name=self.COMPRESSION_LZMA,
                encoder_type=UniModuleDefinition.parse("lzma:compress"),
                decoder_type=UniModuleDefinition.parse("lzma:decompress"),
                dynamic_props_={}
            ),
        }

        for name, definition, other_props in self._parse_definition('compression', config.get("compression", dict()), dict(), {"encoder_import_template", "decoder_import_template"}):
            result[name] = UniCodecDefinition(
                name=name,
                encoder_type=UniModuleDefinition.parse(self._util.template.template(definition['encoder_import_template'], name=name)),
                decoder_type=UniModuleDefinition.parse(self._util.template.template(definition['decoder_import_template'], name=name)),
                dynamic_props_={}
            )

        return result

    def _parse_codecs(self, config: Dict[str, Any]) -> Dict[str, UniCodecDefinition]:
        result = {
            self.APPLICATION_JSON: UniCodecDefinition(
                name=self.APPLICATION_JSON,
                encoder_type=UniModuleDefinition.parse("unipipeline.utils.complex_serializer:complex_serializer_json_dumps"),
                decoder_type=UniModuleDefinition.parse("json:loads"),
                dynamic_props_={}
            ),
        }

        for name, definition, other_props in self._parse_definition('codecs', config.get("codecs", dict()), dict(), {"encoder_import_template", "decoder_import_template"}):
            result[name] = UniCodecDefinition(
                name=name,
                encoder_type=UniModuleDefinition.parse(self._util.template.template(definition['encoder_import_template'], name=name)),
                decoder_type=UniModuleDefinition.parse(self._util.template.template(definition['decoder_import_template'], name=name)),
                dynamic_props_={}
            )

        return result

    def _parse_cron_tasks(self, config: Dict[str, Any], service: UniServiceDefinition, workers: Dict[str, UniWorkerDefinition]) -> Dict[str, UniCronTaskDefinition]:
        result = dict()
        defaults = dict(
            alone=True,
            every_sec=None,
            when=None,
        )
        for name, definition, other_props in self._parse_definition("cron", config.get("cron", dict()), defaults, {"worker"}):
            worker_def = workers[definition["worker"]]
            if worker_def.input_message.name != UNI_CRON_MESSAGE:
                raise ValueError(f"input_message of worker '{worker_def.name}' must be '{UNI_CRON_MESSAGE}'. '{worker_def.input_message.name}' was given")
            result[name] = UniCronTaskDefinition(
                id=definition["id"],
                name=name,
                worker=worker_def,
                every_sec=definition.get('every_sec', None),
                when=definition.get('when', None),
                alone=definition["alone"],
                dynamic_props_=other_props,
            )
        return result

    def _parse_messages(self, config: Dict[str, Any], service: UniServiceDefinition) -> Dict[str, UniMessageDefinition]:
        result = {
            UNI_CRON_MESSAGE: UniMessageDefinition(
                name=UNI_CRON_MESSAGE,
                type=UniModuleDefinition(
                    module="unipipeline.message.uni_cron_message",
                    object_name="UniCronMessage",
                ),
                dynamic_props_=dict(),
            )
        }

        if "messages" not in config:
            raise UniConfigError('messages is not defined in config')

        for name, definition, other_props in self._parse_definition("messages", config["messages"], dict(), {"import_template", }):
            import_template = definition.pop("import_template")
            id_ = definition.pop("id")
            result[name] = UniMessageDefinition(
                **definition,
                type=UniModuleDefinition.parse(self._util.template.template(import_template, **definition, **{"service": service, "id": id_})),
                dynamic_props_=other_props,
            )

        return result

    def _parse_waitings(self, config: Dict[str, Any], service: UniServiceDefinition) -> Dict[str, UniWaitingDefinition]:
        result = dict()
        defaults = dict(
            retry_max_count=3,
            retry_delay_s=10,
        )

        for name, definition, other_props in self._parse_definition('waitings', config.get('waitings', dict()), defaults, {"import_template", }):
            import_template = definition.pop("import_template")
            result[name] = UniWaitingDefinition(
                **definition,
                type=UniModuleDefinition.parse(self._util.template.template(import_template, **definition, **{"service": service})),
                dynamic_props_=other_props
            )

        return result

    def _parse_brokers(self, config: Dict[str, Any], service: UniServiceDefinition, external: Dict[str, UniExternalDefinition]) -> Dict[str, UniBrokerDefinition]:
        result: Dict[str, UniBrokerDefinition] = dict()
        defaults = dict(
            retry_max_count=3,
            retry_delay_s=10,

            content_type=self.APPLICATION_JSON,
            compression=None,
            external=None,
        )

        if "brokers" not in config:
            raise UniConfigError('brokers is not defined in config')

        for name, definition, other_def in self._parse_definition("brokers", config["brokers"], defaults, {"import_template", }):
            ext = definition["external"]
            if ext is not None and ext not in external:
                raise UniConfigError(f'worker_definition brokers->{name} has invalid external: "{ext}"')

            import_template = definition.pop('import_template')
            result[name] = UniBrokerDefinition(
                **definition,
                type=UniModuleDefinition.parse(self._util.template.template(import_template, **definition, **{"service": service})),
                dynamic_props_=other_def,
            )
        return result

    def _parse_service(self, config: Dict[str, Any]) -> UniServiceDefinition:
        if "service" not in config:
            raise UniConfigError('service is not defined in config')

        service_conf = config["service"]

        if "name" not in service_conf:
            raise UniConfigError('service->name is not defined')

        clrs = service_conf.get('echo_colors', True)
        lvl = service_conf.get('echo_level', 'warning')

        self._echo.level = lvl  # TODO: enable verbose if set
        self._util.color.enabled = clrs

        return UniServiceDefinition(
            name=service_conf["name"],
            id=uuid4(),
            colors=clrs,
            echo_level=lvl,
        )

    def _parse_external_services(self, config: Dict[str, Any]) -> Dict[str, UniExternalDefinition]:
        if "external" not in config:
            return dict()

        external_conf = config["external"]

        defaults: Dict[str, Any] = dict()

        result = dict()

        for name, definition, other_props in self._parse_definition('external', external_conf, defaults, set()):
            if other_props:
                raise UniConfigError(f'external->{name} has invalid props: {set(other_props.keys())}')

            dfn = UniExternalDefinition(
                **definition,
                dynamic_props_=dict(),
            )
            result[name] = dfn

        return result

    def _parse_workers(
        self,
        config: Dict[str, Any],
        service: UniServiceDefinition,
        brokers: Dict[str, UniBrokerDefinition],
        messages: Dict[str, UniMessageDefinition],
        waitings: Dict[str, UniWaitingDefinition],
        external: Dict[str, UniExternalDefinition],
    ) -> Dict[str, UniWorkerDefinition]:
        result = dict()

        out_workers = set()

        defaults: Dict[str, Any] = dict(
            topic="{{name}}",
            error_payload_topic="{{topic}}__error__payload",
            error_topic="{{topic}}__error",
            answer_topic="{{name}}__answer",
            broker="default_broker",
            external=None,
            answer_message=None,
            prefetch_count=1,

            input_unwrapped=False,
            answer_unwrapped=False,

            # notification_file="/var/unipipeline/{{service.name}}/{{service.id}}/worker_{{name}}_{{id}}/metrics",

            waiting_for=[],
            output_workers=[],
        )

        if "workers" not in config:
            raise UniConfigError('workers is not defined in config')

        for name, definition, other_props in self._parse_definition("workers", config["workers"], defaults, {"import_template", "input_message"}):
            for ow in definition["output_workers"]:
                out_workers.add(ow)

            br = definition["broker"]
            if br not in brokers:
                raise UniConfigError(f'worker_definition workers->{name} has invalid broker: {br}')
            definition["broker"] = brokers[br]

            im = definition["input_message"]
            if im not in messages:
                raise UniConfigError(f'worker_definition workers->{name} has invalid input_message: {im}')
            definition["input_message"] = messages[im]

            om = definition['answer_message']
            if om is not None:
                if om not in messages:
                    raise UniConfigError(f'worker_definition workers->{name} has invalid answer_message: {om}')
                definition["answer_message"] = messages[om]

            ext = definition["external"]
            if ext is not None and ext not in external:
                raise UniConfigError(f'worker_definition workers->{name} has invalid external: "{ext}"')

            waitings_: Set[UniWaitingDefinition] = set()
            for w in definition.pop('waiting_for'):
                if w not in waitings:
                    raise UniConfigError(f'worker_definition workers->{name} has invalid waiting_for: {w}')
                waitings_.add(waitings[w])

            topic_template = definition.pop('topic')
            error_topic_template = definition.pop('error_topic')
            error_payload_topic_template = definition.pop('error_payload_topic')
            answer_topic_template = definition.pop('answer_topic')

            topic_templates = {topic_template, error_topic_template, error_payload_topic_template, answer_topic_template}
            if len(topic_templates) != 4:
                raise UniConfigError(f'worker_definition workers->{name} has duplicate topic templates: {", ".join(topic_templates)}')

            template_data: Dict[str, Any] = {**definition, "service": service}
            topic = self._util.template.template(topic_template, **template_data)
            template_data['topic'] = topic

            import_template = definition.pop("import_template")

            defn = UniWorkerDefinition(
                **definition,
                type=UniModuleDefinition.parse(self._util.template.template(import_template, **template_data)) if definition["external"] is None else None,
                topic=topic,
                error_topic=self._util.template.template(error_topic_template, **template_data),
                error_payload_topic=self._util.template.template(error_payload_topic_template, **template_data),
                answer_topic=self._util.template.template(answer_topic_template, **template_data),
                waitings=waitings_,
                dynamic_props_=other_props,
            )

            result[name] = defn

        out_intersection_workers = set(result.keys()).intersection(out_workers)
        if len(out_intersection_workers) != len(out_workers):
            raise UniConfigError(f'workers worker_definition has invalid worker_names (in output_workers prop): {", ".join(out_intersection_workers)}')

        return result



from enum import Enum


class UniMessageMetaErrTopic(Enum):
    SYSTEM_ERR = 'error_system'
    MESSAGE_PAYLOAD_ERR = 'error_message_payload'
    ANSWER_MESSAGE_PAYLOAD_ERR = 'error_answer_message_payload'
    HANDLE_MESSAGE_ERR = 'error_message_handling'
    ERROR_HANDLING_ERR = 'error_handling'



from unipipeline.brokers.uni_amqp_py_broker import UniAmqpPyBroker


class RmqBroker(UniAmqpPyBroker):

    @classmethod
    def get_connection_uri(cls) -> str:
        return 'amqp://admin:admin@localhost:25672'



from atexit import register as on_exit_app
from typing import Set, List, TYPE_CHECKING, Generic, TypeVar, Type, Optional, Dict, Any

from pydantic import ValidationError

from unipipeline.brokers.uni_broker_consumer import UniBrokerConsumer
from unipipeline.definitions.uni_broker_definition import UniBrokerDefinition
from unipipeline.definitions.uni_definition import UniDynamicDefinition
from unipipeline.message_meta.uni_message_meta import UniMessageMeta, UniAnswerParams
from unipipeline.utils.uni_echo import UniEcho

if TYPE_CHECKING:
    from unipipeline.modules.uni_mediator import UniMediator


TConf = TypeVar('TConf')


class UniBroker(Generic[TConf]):
    config_type: Type[TConf] = UniDynamicDefinition  # type: ignore

    def __init__(self, mediator: 'UniMediator', definition: UniBrokerDefinition) -> None:
        self._uni_definition = definition
        self._uni_mediator = mediator
        self._uni_echo = self._uni_mediator.echo.mk_child(f'broker[{self._uni_definition.name}]')

        try:
            self._uni_conf = self._uni_definition.configure_dynamic(self.config_type)  # type: ignore
        except ValidationError as e:
            self._uni_echo.exit_with_error(str(e))

        on_exit_app(self.close)

    @property
    def config(self) -> TConf:
        return self._uni_conf

    def serialize_message_body(self, meta: UniMessageMeta) -> bytes:
        data_dict: Dict[str, Any] = meta.payload if meta.unwrapped else meta.dict()
        data_bytes = self._uni_mediator.serialize_content_type(self.definition.content_type, data_dict)
        data_bytes = self._uni_mediator.compress_message_body(self.definition.compression, data_bytes)
        return data_bytes

    def parse_message_body(self, content: bytes, compression: Optional[str], content_type: str, unwrapped: bool) -> UniMessageMeta:
        content = self._uni_mediator.decompress_message_body(compression, content)
        fields = self._uni_mediator.parse_content_type(content_type, content)
        if unwrapped:
            return UniMessageMeta.create_new(fields, unwrapped=True)
        return UniMessageMeta(**fields)

    def connect(self) -> None:
        raise NotImplementedError(f'method connect must be implemented for {type(self).__name__}')

    def close(self) -> None:
        raise NotImplementedError(f'method close must be implemented for {type(self).__name__}')

    def add_consumer(self, consumer: UniBrokerConsumer) -> None:
        raise NotImplementedError(f'method consume must be implemented for {type(self).__name__}')

    def stop_consuming(self) -> None:
        raise NotImplementedError(f'method stop_consuming must be implemented for {type(self).__name__}')

    def start_consuming(self) -> None:
        raise NotImplementedError(f'method start_consuming must be implemented for {type(self).__name__}')

    def publish(self, topic: str, meta_list: List[UniMessageMeta], alone: bool = False) -> None:
        raise NotImplementedError(f'method publish must be implemented for {type(self).__name__}')

    def rpc_call(self, topic: str, meta: UniMessageMeta, *, alone: bool = False, max_delay_s: int = 1, unwrapped: bool = False) -> Optional[UniMessageMeta]:
        raise NotImplementedError(f'method rpc_call must be implemented for {type(self).__name__}')

    def get_topic_approximate_messages_count(self, topic: str) -> int:
        raise NotImplementedError(f'method get_topic_size must be implemented for {type(self).__name__}')

    def initialize(self, topics: Set[str], answer_topics: Set[str]) -> None:
        raise NotImplementedError(f'method initialize must be implemented for {type(self).__name__}')

    def publish_answer(self, answer_params: UniAnswerParams, meta: UniMessageMeta) -> None:
        raise NotImplementedError(f'method publish_answer must be implemented for {type(self).__name__}')

    @property
    def definition(self) -> UniBrokerDefinition:
        return self._uni_definition

    @property
    def echo(self) -> UniEcho:
        return self._uni_echo



import os.path
import sys
from datetime import datetime
from time import sleep

CWD = os.path.dirname(os.path.abspath(__file__))

sys.path.append(os.path.dirname(CWD))

from example.args import args
from unipipeline.modules.uni import Uni

u = Uni(f"{CWD}/dag-{args.type}.yml")

u.init_producer_worker('input_worker')
u.init_producer_worker('auto_retry_worker')
u.initialize()

for i in range(args.produce_count):
    batch = []
    for j in range(args.produce_batch_count):
        some = "Привет World"  # * 1000000
        batch.append(dict(value=i, some=some))
    u.send_to("input_worker", batch)
    # if i % 1000 == 0:
    u.send_to("auto_retry_worker", batch)
    #     sleep(14)
    print('>> SENT at', datetime.now())  # noqa
    sleep(args.delay)

# u.exit()



from time import sleep
from uuid import UUID

from unipipeline.definitions.uni_definition import UniDefinition
from unipipeline.definitions.uni_module_definition import UniModuleDefinition
from unipipeline.utils.uni_echo import UniEcho
from unipipeline.waiting.uni_wating import UniWaiting


class UniWaitingDefinition(UniDefinition):
    id: UUID
    name: str
    retry_max_count: int
    retry_delay_s: int
    type: UniModuleDefinition

    def __hash__(self) -> int:
        return hash(self.id)

    def wait(self, echo: UniEcho) -> None:
        waiting_type = self.type.import_class(UniWaiting, echo)
        for try_count in range(self.retry_max_count):
            try:
                w = waiting_type()
                w.try_to_connect()
                echo.log_info(f'access to {self.name} was reached')
                return
            except ConnectionError:
                echo.log_debug(f'retry wait for {self.name} [{try_count}/{self.retry_max_count}]')
                sleep(self.retry_delay_s)
                continue
        raise ConnectionError('unavailable connection to %s', waiting_type.__name__)



from uuid import UUID

from pydantic import BaseModel, Extra


class UniAnswerParams(BaseModel):
    topic: str
    id: UUID
    ttl_s: int

    class Config:
        frozen = True
        extra = Extra.ignore
